
#include <time.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

/* compute high precision walltime and walltime difference */
static double wallclock(const double * ref)
{
    struct timespec t;
    double ret;

    clock_gettime(CLOCK_REALTIME, &t);
    ret = ((double) t.tv_sec)*1.0e6 + 1.0e-3*((double) t.tv_nsec);

    return ref ? (ret - *ref) : ret;
}


/* implement a simple vector addition: C = A + B */

void vector_add_noopt(float *a, float *b, float *c, int dim)
{
    int i;

    for (i = 0; i < dim; ++i) {
        c[i] = a[i] + b[i];
    }
}

void vector_add_unroll8(const float * __restrict a, const float * __restrict b,
                        float * __restrict c, int dim)
{
    int i,pre;

#if defined(__GNUC__) && !defined(__INTEL_COMPILER)
    const float *x = __builtin_assume_aligned(a,64);
    const float *y = __builtin_assume_aligned(b,64);
    float *z = __builtin_assume_aligned(c,64);
#else
    const float *x = a;
    const float *y = b;
    float *z = c;
#endif

    pre = dim & 8;

    for (i = 0; i < pre; ++i) {
        z[i] = x[i] + y[i];
    }

    for (i = pre; i < dim; i +=8) {
        z[i] = x[i] + y[i];
        z[i+1] = x[i+1] + y[i+1];
        z[i+2] = x[i+2] + y[i+2];
        z[i+3] = x[i+3] + y[i+3];
        z[i+4] = x[i+4] + y[i+4];
        z[i+5] = x[i+5] + y[i+5];
        z[i+6] = x[i+6] + y[i+6];
        z[i+7] = x[i+7] + y[i+7];
    }
}

int main(int argc, char **argv)
{
    double secs;
    float *a, *b, *c;
    int dim,i,flag;

    if (argc < 2) {
        printf("usage: %s <dim>\n", argv[0]);
        return 1;
    }

    dim = atoi(argv[1]);

    if (dim < 1) {
        printf("usage: %s <dim>, with dim > 0\n", argv[0]);
        return 2;
    }

    posix_memalign((void **)&a,64,dim*sizeof(float));
    posix_memalign((void **)&b,64,dim*sizeof(float));
    posix_memalign((void **)&c,64,dim*sizeof(float));

    /* fill vectors */
    for (i=0; i < dim; ++i) {
        a[i] = i;
        b[i] = dim - i;
    }
    /* warm up cache */
    vector_add_noopt(a,b,c,dim);
    vector_add_noopt(a,b,c,dim);

    /* run on CPU and measure time */
    secs = wallclock(NULL);
    vector_add_noopt(a,b,c,dim);
    printf("time for noopt code: %g us\n",wallclock(&secs));
    
    /* check result */
    flag=0;
    for (i=0; i < dim; ++i) {
        if (c[i] != dim) {
            flag=1;
        }
    }
    if (flag != 0) {
        printf("noopt check FAIL\n");
    }

    /* run 8x unroll version and measure time */
    secs = wallclock(NULL);
    vector_add_unroll8(a,b,c,dim);
    printf("time for 8x unrolled code: %g us\n",wallclock(&secs));
    
    /* check result */
    flag=0;
    for (i=0; i < dim; ++i) {
        if (c[i] != dim) {
            flag=1;
        }
    }
    if (flag != 0) {
        printf("8x unrolled check FAIL\n");
    }

    return 0;
}
